;============================================
; Library module source file:
;	GetOurConfig()
;	FreeConfigMem()
;============================================


	INCLUDE	/masm32/include/masm32rt.inc
	INCLUDE	config.inc


;============================================
; Defines, macros, prototypes, etc.
;============================================

; Provide storage for ($maxKeys) entries, each ($avgKeyEntryLength) long (+ SIZEOF $key for table entry)
$maxKeys		EQU 100
$avgKeyEntryLength	EQU 25
$keyStringSpace		EQU $maxKeys * $avgKeyEntryLength
$keyHeapSize		EQU ($maxKeys * SIZEOF $key) + $keyStringSpace

$cr		EQU 13
$lf		EQU 10
$EOF		EQU 0FFH		;Our own made-up signal value for end o'file.


;============================================
; HERE BE DATA
;============================================
.data

HeapHandle		HANDLE NULL
FileHeapPtr		DD NULL
KeyHeapPtr		DD NULL

HeapTablePtr		DD ?
HeapStringPtr		DD ?



GCparseTable		LABEL DWORD
;	  sp    ;     =    EOL   EOF  <any>
  DD	_pn0, _pn1, _pnE, _pn0, _pnF, _pnA	;row 0
  DD	_pn1, _pn1, _pn1, _pn0, _pnF, _pn1	;row 1
  DD	_pn4, _pnE, _pn3, _pnE, _pnE, _pn2	;row 2
  DD	_pn3, _pnD, _pnE, _pnC, _pnB, _pn5	;row 3
  DD	_pn4, _pnE, _pn3, _pnE, _pnE, _pnE	;row 4
  DD	_pn6, _pnD, _pnE, _pnC, _pnB, _pn5	;row 5
  DD	_pn6, _pnE, _pnE, _pnC, _pnB, _pnE	;row 6

$GCrow1			EQU ($numGCparseChars + 1) * 4
$GCrow2			EQU $GCrow1 * 2
$GCrow3			EQU $GCrow1 * 3
$GCrow4			EQU $GCrow1 * 4
$GCrow5			EQU $GCrow1 * 5
$GCrow6			EQU $GCrow1 * 6

GCparseChars		DB ' ', ';', '=', $cr, $EOF
$numGCparseChars	EQU $ - GCparseChars
$anyCharOffset		EQU $numGCparseChars * 4
ConfigSuffixStr		DB ".cfg", 0


;============================================
; CODE LIVES HERE
;============================================
.code


;============================================
; GetOurConfig (progName, keyPtr)
;
; Opens and reads configuration fie if it exists. Filename is assumed to be "<progName>.cfg",
; If file is read successfully, key/value data structure is filled in.
;
; Parameters:
;
; progName
;	A pointer to a NULL-terminated string with the program's name.
; keyPtr
;	A pointer to a variable (DWORD) that receives the address of the returned key table
;	(the table is kept in static storage until freed by FreeConfigMem() ).
;	The table consists of $key structures (defined in include file).
;	The end of the table is marked by a "key" value of -1.
;	"Val" strings can be NULL (consisting only of the terminating zero).
;
; Return values:
;	0: config file successfully read
;	otherwise an error condition:
;		GOC_FILE_NOT_FOUND
;		GOC_READ_ERROR
;		GOC_FILE_TOO_BIG
;		GOC_PARSE_ERROR
;		GOC_MEMORY_ERROR
;============================================

GetOurConfig	PROC	progName:DWORD, keyTablePtr:DWORD
		LOCAL	cfgHandle:HANDLE, fileSize:DWORD, numBytesRead:DWORD
		LOCAL	buffer1[100]:BYTE, errCode:DWORD, readCount:DWORD
		LOCAL	keyPtr:DWORD, valPtr:DWORD, heapCount:DWORD

	PUSH	EBX
	PUSH	ESI
	PUSH	EDI

; Create "<progname>.cfg" string:
	MOV	ESI, progName
	LEA	EDI, buffer1
	PUSH	EDI
	CALL	_strcpy
	LEA	ESI, ConfigSuffixStr
	POP	EDI
	CALL	_strcat
	INVOKE	CreateFile, ADDR buffer1, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING,
		FILE_ATTRIBUTE_NORMAL, NULL
	CMP	EAX, INVALID_HANDLE_VALUE
	JNE	goc10
	MOV	EAX, GOC_READ_ERROR
; If file heap is still allocated, free it:
goc99:	CMP	FileHeapPtr, NULL
	JE	goc999
	PUSH	EAX
	INVOKE	HeapFree, HeapHandle, 0, FileHeapPtr
	MOV	FileHeapPtr, NULL
	POP	EAX
goc999:	POP	EDI
	POP	ESI
	PoP	EBX	
	RET

goc10:	MOV	cfgHandle, EAX
	INVOKE	GetFileSize, EAX, NULL
	MOV	fileSize, EAX
	CMP	fileSize, GOC_MAX_FILESIZE
	JBE	goc20
	MOV	errCode, GOC_FILE_TOO_BIG
filecloseexit:
	INVOKE	CloseHandle, cfgHandle
	MOV	EAX, errCode
	JMP	goc99

; File open, size gotten, ready to be read: Must allocate heap:
goc20:	CMP	HeapHandle, NULL		;Has heap handle been gotten?
	JNE	goc25
; Get handle to heap memory to allocate later:
	INVOKE	GetProcessHeap
	MOV	HeapHandle, EAX
goc25:	INVOKE	HeapAlloc, HeapHandle, 0,  fileSize
	OR	EAX, EAX			;Check that allocation succeeded.
	JNZ	goc30
	MOV	errCode, GOC_MEMORY_ERROR
	JMP	filecloseexit

; Read file:
goc30:	MOV	FileHeapPtr, EAX
	INVOKE	ReadFile, cfgHandle, FileHeapPtr, fileSize, ADDR numBytesRead, NULL
	OR	EAX, EAX
	JNZ	goc35
	MOV	errCode, GOC_READ_ERROR
freememexit:
	INVOKE	HeapFree, HeapHandle, 0, FileHeapPtr
	MOV	FileHeapPtr, NULL
	JMP	filecloseexit

;============================================
; File parsing starts here
;============================================

goc35:	MOV	EAX, numBytesRead
	MOV	readCount, EAX
	INVOKE	HeapAlloc,  HeapHandle, 0, $keyHeapSize
	MOV	KeyHeapPtr, EAX

; Set up heap pointers:
	MOV	HeapTablePtr, EAX
	LEA	EAX, [EAX + ($maxKeys * SIZEOF $key)]
	MOV	HeapStringPtr, EAX
	MOV	heapCount, 0

	MOV	ESI, FileHeapPtr
	XOR	EBX, EBX			;EBX is row pointer.
parse0:	CMP	readCount, 0			;Check for EOF.
	JNE	p1
	MOV	AL, $EOF
	JMP	SHORT p2
p1:	LODSB
	DEC	readCount
	CMP	AL, $lf
	JE	parse0				;Scan off line-feeds.
p2:	LEA	EDI, GCparseChars
	MOV	EDX, EDI
	MOV	ECX, $numGCparseChars
	REPNE	SCASB
	JNE	parse10				;No match = <any>.
	DEC	EDI
	SUB	EDI,  EDX			;Get offset of match.
	SHL	EDI, 2				;Convert to DWORD offset.
	JMP	[EDI + EBX + GCparseTable]	;Jump to next node.

parse10:
	JMP	[EBX + GCparseTable + $anyCharOffset]


;============================================
; PARSING NODES
;============================================

_pn0	LABEL NEAR
	XOR	EBX, EBX
	JMP	parse0

_pn1	LABEL NEAR
	MOV	EBX, $GCrow1
	JMP	parse0

_pnA	LABEL NEAR
; Set up <key> storage:
	MOV	EDX, HeapStringPtr
	MOV	keyPtr, EDX
; Fall through to _pn2:

_pn2	LABEL NEAR
; Store <key> character:
	INC	heapCount
	CMP	heapCount, $keyHeapSize
	JA	heapSizeError
	MOV	EDI,  HeapStringPtr
	STOSB
	MOV	HeapStringPtr, EDI		;*****
	MOV	EBX, $GCrow2
	JMP	parse0

_pn3	LABEL NEAR
	INC	heapCount
	CMP	heapCount, $keyHeapSize
	JA	heapSizeError
; Terminate <key> string:
	MOV	EDX,  HeapStringPtr
	MOV	BYTE PTR [EDX], NULL
	INC	HeapStringPtr			;*****
; Set up <val> storage;
	MOV	EDX, HeapStringPtr
	MOV	valPtr, EDX
	MOV	EBX, $GCrow3
	JMP	parse0

_pn4	LABEL NEAR
	MOV	EBX, $GCrow4
	JMP	parse0

_pnB	LABEL NEAR
	MOV	EDX, OFFSET  _pnF		;Jump destination.
; Fall throught to common code:

; Common code here (nodes B, C & D):
_pnB_C_D:
	ADD	heapCount, SIZEOF  $key + 1	;Incl. NULL byte about to be stored.
	CMP	heapCount, $keyHeapSize
	JA	heapSizeError
; Terminate <val> string:
	MOV	EDI,  HeapStringPtr
	MOV	BYTE PTR [EDI], NULL
	INC	HeapStringPtr			;*****
; Complete current key entry:
	MOV	EDI, HeapTablePtr
	MOV	EAX, keyPtr
	MOV	[EDI + $key.key], EAX
	MOV	EAX, valPtr
	MOV	[EDI + $key.val], EAX
	ADD	HeapTablePtr, SIZEOF $key	;*****
	JMP	EDX				;"Programmed GOTO" (node 0, 1 or F)
	
heapSizeError:
; Heap used up: exit w/error:
	MOV	EAX, GOC_FILE_TOO_BIG
	JMP	goc999

_pn5	LABEL NEAR
	INC	heapCount
	CMP	heapCount, $keyHeapSize
	JA	heapSizeError
; Store <val> character:
	MOV	EDI, HeapStringPtr
	STOSB					;*****
	MOV	HeapStringPtr, EDI
	MOV	EBX, $GCrow5
	JMP	parse0

_pn6	LABEL NEAR
	MOV	EBX, $GCrow6
	JMP	parse0

_pnC	LABEL NEAR
	MOV	EDX, OFFSET _pn0		;Jump destination
	JMP	_pnB_C_D

_pnD	LABEL NEAR
	MOV	EDX, OFFSET _pn1		;Jump destination
	JMP	_pnB_C_D

_pnF	LABEL NEAR
; Final processing, exit w/no error:
	MOV	EDI, HeapTablePtr
	MOV	[EDI + $key.key], -1		;Mark end of table.
	MOV	EDI, keyTablePtr
	MOV	EAX, KeyHeapPtr
	MOV	[EDI], EAX			;Return pointer to table.
	XOR	EAX, EAX
	JMP	goc99

_pnE	LABEL NEAR
; Exit w/syntax error:
	MOV	EAX, GOC_PARSE_ERROR
	JMP	goc99

GetOurConfig	ENDP



;====================================================================
; FreeConfigMem()
;
; Frees any allocated heap used by config. package
; Be sure to call this before your program exits.
;
;====================================================================

FreeConfigMem	PROC

	CMP	FileHeapPtr, NULL
	JE	fcm10
	INVOKE	HeapFree, HeapHandle, 0, FileHeapPtr
	MOV	FileHeapPtr, NULL
fcm10:	CMP	KeyHeapPtr, NULL
	JE	fcm99
	INVOKE	HeapFree, HeapHandle, 0, KeyHeapPtr
	MOV	KeyHeapPtr, NULL
fcm99:	RET

FreeConfigMem	ENDP



;====================================================================
; _strcpy()
;
; On entry,
;   ESI--> string to be copied
;   EDI--> destination buffer
;====================================================================

_strcpy		PROC

scp10:	MOV	AL, BYTE PTR [ESI]
	MOV	[EDI], BYTE PTR AL
	OR	AL, AL				;Stop on NULL.
	JZ	scp99
	INC	ESI
	INC	EDI
	JMP	scp10

scp99:	RET

_strcpy		ENDP


;====================================================================
; _strcat()
;
; On entry,
;   ESI--> string to be added to buffer
;   EDI--> destination buffer
;====================================================================

_strcat		PROC

; Locate end of destination buffer:
scat10:	CMP	BYTE PTR [EDI], 0		;Look for NULL terminator.
	JE	scat20
	LEA	EDI, [EDI + 1]
	JMP	scat10
scat20:	CALL	_strcpy

	RET

_strcat		ENDP


	END
